#include "date.h"

#define Eo(x) { cerr << #x << " = " << (x) << endl; }

namespace {

	template<class T>
		inline T abs(T value) { return value < 0 ? -value : value; }

	string int2String(int i) {
		string res;
		for(; i > 0; i /= 10) res = char(i%10+'0') + res;
		while(res.length() < 2) res = '0' + res;
		return res;
	}

	string month2string(int m) {
		switch(m) {
			case Date::Jan: return "Jan";
			case Date::Feb: return "Feb";
			case Date::Mar: return "Mar";
			case Date::Apr: return "Apr";
			case Date::May: return "May";
			case Date::Jun: return "Jun";
			case Date::Jul: return "Jul";
			case Date::Aug: return "Aug";
			case Date::Sep: return "Sep";
			case Date::Oct: return "Oct";
			case Date::Nov: return "Nov";
			case Date::Dec: return "Dec";
		}
		return "unknown_month";
	}
	
	void replaceString(string& s, string what, string on) {
		size_t lastFind = 0;
		while((lastFind = s.find(what, lastFind)) != s.npos) {
			s.replace(lastFind, what.length(), on);
			lastFind += on.length();
		}
	}
}

namespace Date {

	void Date::normDate() {
		if(second < 0) {
			int tmp = abs(second)/60 + 1;
			second += tmp*60;
			minute -= tmp;
		}
		if(second >= 60) {
			int tmp = second/60;
			second -= tmp*60;
			minute += tmp;
		}

		if(minute < 0) {
			int tmp = abs(minute)/60 + 1;
			minute += tmp*60;
			hour -= tmp;
		}
		if(minute >= 60) {
			int tmp = minute/60;
			minute -= tmp*60;
			hour += tmp;
		}

		if(hour < 0) {
			int tmp = abs(hour)/24 + 1;
			hour += tmp*24;
			day -= tmp;
		}
		if(hour >= 24) {
			int tmp = hour/24;
			hour -= tmp*24;
			day += tmp;
		}

		int64 daysInForYears = (365*4+1)*100-3;
		int64 daysInAllYears = daysInForYears*25;
		int64 d = day;
		if(d < 0) {
			int tmp = abs(day) / daysInAllYears + 1;
			d += daysInAllYears*tmp;
		}
		int m = month;
		m--;
		if(m < 0) {
			int tmp = abs(m)/12 + 1;
			m += tmp*12;
			year -= tmp;
		}
		if(m >= 12) {
			int tmp = m/12;
			m -= tmp*12;
			year += tmp;
		}
		m++;
		if(year < 0) {
			int tmp = abs(year)/10000 + 1;
			year += tmp*10000;
		}
		year %= 10000;
		d %= daysInAllYears;

		const int monthDays[13] = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
		year += (d / daysInForYears)*400;
		d %= daysInForYears;
		while(true) {
			int daysInCurMonth = monthDays[m];
			if(m == 2 && (year % 400 == 0 || (year % 4 == 0 && year % 100 != 0))) daysInCurMonth++;
			if(d <= daysInCurMonth) break;
			d -= daysInCurMonth;
			m++;
			if(m == 13) {
				m = 1;
				year++;
			}
		}
		day = d;
		month = static_cast<Month>(m);
		year %= 10000;
	}

	Date::Date() {
		time_t curtime;
		time(&curtime);
		tm *curutc;
		curutc = gmtime(&curtime);

		year = curutc->tm_year + 1900;
		month = static_cast<Month>(curutc->tm_mon+1);
		day = curutc->tm_mday;
		hour = curutc->tm_hour;
		minute = curutc->tm_min;
		second = curutc->tm_sec;
		//Eo(year); Eo(month); Eo(day); Eo(hour); Eo(minute); Eo(second);
	}

	Date::Date(const Date& other) : year(other.year),
									month(other.month),
									day(other.day),
									hour(other.hour),
									minute(other.minute),
									second(other.second)
	{ }
	
	Date::Date(const int _year, const Month _month, const int _day, const int _hour, const int _minute, const int _second) : year(_year),
																													   month(_month),
																													   day(_day),
																													   hour(_hour),
																													   minute(_minute),
																													   second(_second)
	{
		normDate();
	}
	
	Date::Date(const int _year, const int _month, const int _day, const int _hour, const int _minute, const int _second) : year(_year),
																													   month(static_cast<Month>(_month)),
																													   day(_day),
																													   hour(_hour),
																													   minute(_minute),
																													   second(_second)
	{
		normDate();
	}

	Date::Date(const int _year, const Month _month, const int _day) : year(_year),
																    month(_month),
																    day(_day),
																    hour(0),
																    minute(0),
																    second(0)
	{
		normDate();
	}

	Date::Date(const int _hour, const int _minute, const int _second) : hour(_hour),
																	    minute(_minute),
																	    second(_second)
	{
		Date curDate;
		year = curDate.year;
		month = curDate.month;
		day = curDate.day;
		normDate();
	}

	Date& Date::operator=(const Date& other) {
		if(this != &other) {
			year = other.year;
			month = other.month;
			day = other.day;
			hour = other.hour;
			minute = other.minute;
			second = other.second;
		}
		return *this;
	}

	bool Date::operator<(const Date& other) const {
		if(year != other.year) return year < other.year;
		if(month != other.month) return month < other.month;
		if(day != other.day) return day < other.day;
		if(hour != other.hour) return hour < other.hour;
		if(minute != other.minute) return minute < other.minute;
		return second < other.second;
	}

	int Date::getYear() const {
		return year;
	}

	int Date::getMonth() const {
		return month;
	}

	int Date::getDay() const {
		return day;
	}

	int Date::getHour() const {
		return hour;
	}

	int Date::getMinute() const {
		return minute;
	}

	int Date::getSecond() const {
		return second;
	}

	Date Date::addYear(int y) const {
		return Date(this->year+y, this->month, this->day, this->hour, this->minute, this->second);
	}

	Date Date::addMonth(int m) const {
		return Date(this->year, static_cast<int>(this->month)+m, this->day, this->hour, this->minute, this->second);
	}

	Date Date::addDay(int d) const {
		return Date(this->year, this->month, this->day+d, this->hour, this->minute, this->second);
	}

	Date Date::addHour(int h) const {
		return Date(this->year, this->month, this->day, this->hour+h, this->minute, this->second);
	}

	Date Date::addMinute(int m) const {
		return Date(this->year, this->month, this->day, this->hour, this->minute+m, this->second);
	}

	Date Date::addSecond(int s) const {
		return Date(this->year, this->month, this->day, this->hour, this->minute, this->second+s);
	}

	DateInterval Date::getInterval(const Date& other) const {
		if(*this < other) {
			DateInterval res = other.getInterval(*this);
			res.invert();
			return res;
		}

		int resYear = year - other.getYear();
		int resMonth = month - other.getMonth();
		int resDay = day - other.getDay();
		int resHour = hour - other.getHour();
		int resMinute = minute - other.getMinute();
		int resSecond = second - other.getSecond();
		if(resSecond < 0) {
			resSecond += 60;
			resMinute--;
		}
		if(resMinute < 0) {
			resMinute += 60;
			resHour--;
		}
		if(resHour < 0) {
			resHour += 24;
			resDay--;
		}
		if(resDay < 0) {
			resDay += 30;
			resMonth--;
		}
		if(resMonth < 0) {
			resMonth += 12;
			resYear--;
		}

		return DateInterval(resYear, resMonth, resDay, resHour, resMonth, resSecond);
	}

	Date Date::addInterval(const DateInterval& interval) const {
		Date res(*this);
		res = res.addSecond(interval.getSecond());
		res = res.addMinute(interval.getMinute());
		res = res.addHour(interval.getHour());
		res = res.addDay(interval.getDay());
		res = res.addMonth(interval.getMonth());
		res = res.addYear(interval.getYear());
		return res;
	}

	string Date::toString(const string& format) const {
		string res = format;
		replaceString(res, "YYYY", int2String(getYear()));
		replaceString(res, "MMM", month2string(getMonth()));
		replaceString(res, "MM", int2String(getMonth()));
		replaceString(res, "DD", int2String(getDay()));
		replaceString(res, "hh", int2String(getHour()));
		replaceString(res, "mm", int2String(getMinute()));
		replaceString(res, "ss", int2String(getSecond()));
		return res;
	}

	std::string Date::toString() const {
		return toString("YYYY-MMM-DD hh:mm:ss");
	}
	
	ostream& operator<<(ostream& o, const Date& d) {
		o << d.toString();
		return o;
	}
}
